function [p,Plast,a,q] = QRlower(A,r)

% Implements inversion algorithm for lower banded matrices
% from Theorem 4.3 in manuscript "Computation of 
% quasiseparable representations of Green matrices".
% The algorithm computes Green generators of the
% inverse matrix. 
% 
% Input: A lower banded matrix to be inverted,
%        r lower bandwidth 
%          (i.e., an upper Hessenberg matrix has r = 1)

% Check size of A
[Nr,Nc] = size(A);
if Nr == Nc
    N = Nr;
else
    disp('Matrix is not square')
    return
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 1: compute generators of UR decomposition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Run algorithm from Theorem 4.1
%
% Here U contains the N-1 matrices U_k. 
[U, R] = URfact(A,r);

% Compute lower Green generators of U* 
% (Lemma 4.2)
pU = zeros(N-1,r);  % i-th row is pU(i)
dU = zeros(N-r,1);  % i-th entry is dU(i)
aU = zeros(r,r,N-1);% i-th matrix is aU(i)
qU = zeros(r,N-r);  % i-th column is qU(i)
% pU(N-r+1) is an rxr matrix and will be defined last
for k = 1:N-r
    Ukstar = U(:,:,k)';
    pU(k,:) = Ukstar(1,1:r);
    dU(k) = Ukstar(1,r+1);
    aU(:,:,k) = Ukstar(2:r+1,1:r);
    qU(:,k) = Ukstar(2:r+1,r+1);
end

for k = N-r+1:N-1
    sizeU = N-k+1;
    Ukstar = U(1:sizeU,1:sizeU,k)';
    pU(k,1:sizeU) = Ukstar(1,:);
    aU(1:sizeU-1,1:sizeU,k) = Ukstar(2:sizeU,:);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 2: compute lower Green generators of inv(A)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% allocate generators
p = zeros(N-1,r);

% 2.1: compute the last p-generator P_{N-r+1}
Ptemp = 1/R(N,N);
for k = N-1:-1:N-r+1
    p(k,1:N-k+1) = (pU(k,1:N-k+1)...
        -R(k,k+1:N)*Ptemp*aU(1:N-k,1:N-k+1,k))/R(k,k);
    Ptemp = [p(k,1:N-k+1); Ptemp*aU(1:N-k,1:N-k+1,k)];
end
Plast = Ptemp;

% 2.2: compute generators for indices from 1 to N-r
q = qU;
a = aU;
for k = N-r:-1:1
    p(k,:) = (pU(k,:)-R(k,k+1:N)*Ptemp*a(:,:,k))/R(k,k);
    Ptemp = [p(k,:); Ptemp*a(:,:,k)];
end
% end of algorithm
end
